////////////////////////////////////////////////////////////////////// 
// Main.cpp 
// (c) 2012-2013 Shuang Feng, Dajiang Liu, Goncalo Abecasis
// 
// This file is distributed as part of the RareMetalWorker source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile RareMetalWorker.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday November 28, 2012
// 

#include "Parameters.h"
#include "Pedigree.h"
#include "math.h"
#include <iostream>
#include "FastFit.h"
#include "TraitTransformations.h"
#include "TransformResiduals.h"
#include <stdio.h>
#include <time.h>
#include "Error.h"
#include "AutoFit.h"
#include "PreMeta.h"
#include "OutputKin.h"
#include "WriteLog.h"
//#include <lapacke.h>

int main(int argc, char ** argv)
{
/*
   double a[5*3] = {1,2,3,4,5,1,3,5,2,4,1,4,2,5,3};
   double b[5*2] = {-10,12,14,16,18,-3,14,12,16,16};
   lapack_int info,m,n,lda,ldb,nrhs;
   int i,j;

   m = 5;
   n = 3;
   nrhs = 2;
   lda = 5;
   ldb = 5;

   info = LAPACKE_dgels(LAPACK_COL_MAJOR,'N',m,n,nrhs,a,lda,b,ldb);

   for(i=0;i<n;i++)
   {
      for(j=0;j<nrhs;j++)
      {
         printf("%lf ",b[i+ldb*j]);
      }
      printf("\n");
   }
*/
   printf("\nRAREMETALWORKER %s -- A Forerunner of RareMetal\n"
          "          (c) 2012-2013 Shuang Feng, Dajiang Liu, Goncalo Abecasis\n\n", VERSION);

clock_t start,end;
   start = clock();
   String pedfile, datfile;
   double tol = 0.0001;

   BEGIN_LONG_PARAMETERS(additional)
      LONG_PARAMETER_GROUP("Input Files")
      LONG_STRINGPARAMETER("ped", &pedfile)
      LONG_STRINGPARAMETER("dat", &datfile)
      LONG_STRINGPARAMETER("vcf", &PreMeta::vcfInput)
      LONG_PARAMETER_GROUP("Output Files")
      LONG_STRINGPARAMETER("prefix", &PreMeta::outputFile)
      LONG_INTPARAMETER("LDwindow", &PreMeta::window)
      LONG_PARAMETER_GROUP("VC Options")
      //LONG_PARAMETER("vcShared", &AutoFit::fitSharedEnvironment)
      LONG_PARAMETER("vcX", &AutoFit::fitX)
      LONG_PARAMETER("useCovariates", &FastFit::useCovariates)
      LONG_PARAMETER_GROUP("Trait Options")
      LONG_PARAMETER("makeResiduals", &FastFit::regressOutCov)
      LONG_PARAMETER("inverseNormal", &FastFit::inverseNormal)
      LONG_STRINGPARAMETER("traitName", &FastFit::traitName)
      LONG_PARAMETER_GROUP("Kinship Source")
      LONG_PARAMETER("kinPedigree", &FastTransform::pedKin)
      LONG_PARAMETER("kinGeno", &FastTransform::empKin)
      LONG_STRINGPARAMETER("kinFile", &FastTransform::readInEmp)
      LONG_PARAMETER("kinSave", &OutputKin::outputKin)
      LONG_PARAMETER_GROUP("Kinship Options")
      LONG_DOUBLEPARAMETER("kinMaf", &KinshipEmp::q)
      LONG_DOUBLEPARAMETER("kinMiss", &KinshipEmp::miss)
      LONG_PARAMETER_GROUP("Chromosome X")
      LONG_STRINGPARAMETER("xLabel", &PreMeta::xLabel)
      LONG_INTPARAMETER("xStart", &PreMeta::Xstart)
      LONG_INTPARAMETER("xEnd", &PreMeta::Xend)
      LONG_INTPARAMETER("maleLabel", &PreMeta::maleLabel)
      LONG_INTPARAMETER("femaleLabel", &PreMeta::femaleLabel)

      END_LONG_PARAMETERS();


   ParameterList pl;
   pl.Add(new LongParameters("Options:", additional));
   pl.Read(argc, argv);

   String logFile;
   if(PreMeta::outputFile != "")
   {
      if(PreMeta::outputFile.FastFindLastChar('.')!=-1 || PreMeta::outputFile.FastFindLastChar('/')!=-1)
	 logFile = PreMeta::outputFile + "singlevar.log";
     else
	 logFile = PreMeta::outputFile + ".singlevar.log";
   }
   else
      logFile = "singlevar.log";

   FILE * log;

   pl.Status();

   printf("Loading PED file ... ");
   log = freopen (logFile,"wt",stderr);
   Pedigree ped;
   ped.Prepare(datfile);
   ped.Load(pedfile);
   printf("done.\n\n");

   if(FastTransform::readInEmp!="")
      FastTransform::empKin=true;

   if(ped.familyCount == ped.count && !FastTransform::empKin && !FastTransform::pedKin && FastTransform::readInEmp=="")
   {
      FastFit::unrelated = true;
      printf("Handling these individuals as unrelated.\n");
   }

   if(ped.markerCount>0)
      PreMeta::genoFromPed = true;
   if(PreMeta::vcfInput !="")
      PreMeta::genoFromVCF=true;
   if(PreMeta::genoFromPed && PreMeta::genoFromVCF)
   {
      PreMeta::genoFromPed=false;
      printf("Warning: Both of your PED and VCF files have genotype information.Genotype will be read from VCF file.\n");
      fprintf(log,"Warning: Both of your PED and VCF files have genotype information. Genotype will be read from VCF file.\n");
   }
   if(PreMeta::dosage && !PreMeta::genoFromVCF)
   {
      fprintf(log,"ERROR! Dosage has to be from vcf file.\n");
      error("ERROR! Dosage has to be from vcf file.\n");
   }
   AutoFit fitEngine;
   FastFit engine;
   FastTransform trans;

   StringArray traits;
   if(!FastFit::traitName.IsEmpty())
   {

      if(FastFit::traitName.Find(".txt",0) !=-1)
	 //if(FastFit::traitName == "traitName.txt")
      {
	 FILE * trFile = fopen(FastFit::traitName,"r");
	 String buffer;
	 while (!feof(trFile))
	 {
	    buffer.ReadLine(trFile);
	    traits.AddTokens(buffer, SEPARATORS);
	 }
	 fclose(trFile);
      }
      else
      {
	 traits.AddTokens(FastFit::traitName,"/");
      }
   }
   else 
   {
      traits = ped.traitNames;
   }

   for(int i=0;i<traits.Length();i++)
   {
      FILE * SCOREoutput;
      FILE * SCOREcov;
      String filename,filename_COV;

      if(PreMeta::outputFile != "")
      {
	 if(PreMeta::outputFile.FastFindLastChar('.')!=-1 || PreMeta::outputFile.FastFindLastChar('/')!=-1)
	 {
	    filename = PreMeta::outputFile + traits[i] + ".singlevar.score.txt";
	    filename_COV = PreMeta::outputFile + traits[i] + ".singlevar.cov.txt";
	 }
	 else
	 {
	    filename = PreMeta::outputFile + "." + traits[i] + ".singlevar.score.txt";
	    filename_COV = PreMeta::outputFile + "." + traits[i] + ".singlevar.cov.txt";
	 }
      }
      else
      {
	 filename_COV = traits[i] + ".singlevar.cov.txt";
	 filename = traits[i] + ".singlevar.score.txt";
      }

      SCOREoutput = fopen(filename,"wt");
      SCOREcov = fopen(filename_COV,"wt");

      String traitName;
      traitName.Copy(traits[i]);
      engine.FitLinearModels(SCOREoutput,fitEngine,ped,tol,trans,traitName,log);

      PreMeta preMeta;
      preMeta.Run(SCOREoutput,SCOREcov,ped,trans,engine,traits[i],log);

      //printf("#GENOMIC CONTROL LAMBDA = %g\n",preMeta.lambda);

      fclose(SCOREoutput);
      fclose(SCOREcov);

      end = clock();
      printf("Association tests for trait %s are finished.\n",traitName.c_str());
      fprintf(log,"Association tests for trait %s are finished.\n",traitName.c_str());
      String scoreFile,covFile;
      printf("Summary statistics for trait %s have been saved in %s.\n",traitName.c_str(),filename.c_str());
      printf("Covariance matrices for trait %s have been saved in %s.\n",traitName.c_str(),filename_COV.c_str());
      fprintf(log,"\nSummary statistics for trait %s have been saved in %s.\n",traitName.c_str(),filename.c_str());
      fprintf(log,"Covariance matrices for trait %s have been saved in %s.\n",traitName.c_str(),filename_COV.c_str());
   }

   WriteLog(pedfile,datfile,log);
   printf("log file has been saved in %s\n\n",logFile.c_str());
   //fprintf(log,"\nlog file has been saved in %s\n\n",logFile.c_str());

   fclose(log);
   //fclose(stderr);
   return 0;
}
